import importlib.util
from urllib.parse import quote_plus
import json

# Widget dependencies
import anywidget
from traitlets import Unicode, Dict, List, Int, Bool
import time
import uuid

# Server dependencies
from uvicorn import Config, Server

from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.middleware.cors import CORSMiddleware
from threading import Thread
import socket

MAX_PORT_TRIES = 1000
DEFAULT_PORT = 8000


class BackgroundServer:
    # Reference: https://github.com/gosling-lang/gos/blob/main/gosling/data/_background_server.py#L10
    def __init__(self, routes):
        middleware = [
            Middleware(CORSMiddleware, allow_origins=[
                       '*'], allow_methods=["OPTIONS", "GET"], allow_headers=['Range'])
        ]
        self.app = Starlette(debug=True, routes=routes, middleware=middleware)
        self.port = None
        self.thread = None
        self.server = None

    def stop(self):
        if self.thread is None:
            return self
        assert self.server is not None
        try:
            self.server.should_exit = True
            self.thread.join()
        finally:
            self.server = None
            self.thread = None
        return self

    def start(self, port=None, timeout=1, daemon=True, log_level="warning"):
        if self.thread is not None:
            return self

        config = Config(
            app=self.app,
            port=port,
            timeout_keep_alive=timeout,
            log_level=log_level
        )
        self.port = config.port
        self.server = Server(config=config)
        self.thread = Thread(target=self.server.run, daemon=daemon)
        self.thread.start()

        while not self.server.started:
            time.sleep(1e-3)

        return self


class VitessceDataServer:
    def __init__(self):
        self.served_configs = []

    def stop_all(self):
        for config in self.served_configs:
            config.stop_all_servers()
        self.served_configs = []

    def register(self, config):
        if config not in self.served_configs:
            self.served_configs.append(config)


# Create a singleton to have a global way to stop all servers
# attached to configs.
data_server = VitessceDataServer()


def is_port_in_use(port):
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(('localhost', port)) == 0


def get_base_url_and_port(port, next_port, proxy=False, base_url=None, host_name=None):
    if port is None:
        use_port = next_port
        next_port += 1
        port_tries = 1
        while is_port_in_use(use_port) and port_tries < MAX_PORT_TRIES:
            use_port = next_port
            next_port += 1
            port_tries += 1
    else:
        use_port = port

    if base_url is None:
        if proxy:
            if importlib.util.find_spec('jupyter_server_proxy') is None:
                raise ValueError(
                    "To use the widget through a proxy, jupyter-server-proxy must be installed.")
            if host_name is None:
                base_url = f"proxy/{use_port}"
            else:
                base_url = f"{host_name}/proxy/{use_port}"
        else:
            base_url = f"http://localhost:{use_port}"

    return base_url, use_port, next_port


def serve_routes(config, routes, use_port):
    if not config.has_server(use_port) and len(routes) > 0:
        server = BackgroundServer(routes)
        config.register_server(use_port, server)
        data_server.register(config)
        server.start(port=use_port)


def launch_vitessce_io(config, theme='light', port=None, base_url=None, host_name=None, proxy=False, open=True):
    import webbrowser
    base_url, use_port, _ = get_base_url_and_port(
        port, DEFAULT_PORT, proxy=proxy, base_url=base_url, host_name=host_name)
    config_dict = config.to_dict(base_url=base_url)
    routes = config.get_routes()
    serve_routes(config, routes, use_port)
    vitessce_url = f"http://vitessce.io/#?theme={theme}&url=data:," + quote_plus(
        json.dumps(config_dict))
    if open:
        webbrowser.open(vitessce_url)
    return vitessce_url


def get_uid_str(uid):
    if uid is None or not str(uid).isalnum():
        uid_str = str(uuid.uuid4())[:4]
    else:
        uid_str = uid
    return uid_str


ESM = """
import { importWithMap } from 'https://unpkg.com/dynamic-importmap@0.1.0';
const importMap = {
  imports: {
    "react": "https://esm.sh/react@18.2.0?dev",
    "react-dom": "https://esm.sh/react-dom@18.2.0?dev",
    "react-dom/client": "https://esm.sh/react-dom@18.2.0/client?dev",
  },
};

const React = await importWithMap("react", importMap);
const { createRoot } = await importWithMap("react-dom/client", importMap);

const e = React.createElement;

const prefersDark = window.matchMedia && window.matchMedia('(prefers-color-scheme: dark)').matches;

// The jupyter server may be running through a proxy,
// which means that the client needs to prepend the part of the URL before /proxy/8000 such as
// https://hub.gke2.mybinder.org/user/vitessce-vitessce-python-swi31vcv/proxy/8000/A/0/cells
function prependBaseUrl(config, proxy, hasHostName) {
  if(!proxy || hasHostName) {
    return config;
  }
  const { origin } = new URL(window.location.href);
  let baseUrl;
  const jupyterLabConfigEl = document.getElementById('jupyter-config-data');

  if (jupyterLabConfigEl) {
    // This is jupyter lab
    baseUrl = JSON.parse(jupyterLabConfigEl.textContent || '').baseUrl;
  } else {
    // This is jupyter notebook
    baseUrl = document.getElementsByTagName('body')[0].getAttribute('data-base-url');
  }
  return {
    ...config,
    datasets: config.datasets.map(d => ({
      ...d,
      files: d.files.map(f => ({
        ...f,
        url: `${origin}${baseUrl}${f.url}`,
      })),
    })),
  };
}

async function render(view) {
    const cssUid = view.model.get('uid');
    const jsDevMode = view.model.get('js_dev_mode');
    const jsPackageVersion = view.model.get('js_package_version');
    const customJsUrl = view.model.get('custom_js_url');
    const pluginEsmArr = view.model.get('plugin_esm');
    const remountOnUidChange = view.model.get('remount_on_uid_change');
    const storeUrls = view.model.get('store_urls');
    const invokeTimeout = view.model.get('invoke_timeout');

    const pkgName = (jsDevMode ? "@vitessce/dev" : "vitessce");

    importMap.imports["vitessce"] = (customJsUrl.length > 0
        ? customJsUrl
        : `https://unpkg.com/${pkgName}@${jsPackageVersion}`
    );

    const {
        Vitessce,
        PluginFileType,
        PluginViewType,
        PluginCoordinationType,
        PluginJointFileType,
        z,
        useCoordination,
        useGridItemSize,
        // TODO: names and function signatures are subject to change for the following functions
        // Reference: https://github.com/keller-mark/use-coordination/issues/37#issuecomment-1946226827
        useComplexCoordination,
        useMultiCoordinationScopesNonNull,
        useMultiCoordinationScopesSecondaryNonNull,
        useComplexCoordinationSecondary,
        useCoordinationScopes,
        useCoordinationScopesBy,
    } = await importWithMap("vitessce", importMap);

    let pluginViewTypes = [];
    let pluginCoordinationTypes = [];
    let pluginFileTypes = [];
    let pluginJointFileTypes = [];

    const stores = Object.fromEntries(
        storeUrls.map(storeUrl => ([
            storeUrl,
            {
                async get(key) {
                    const [data, buffers] = await view.experimental.invoke("_zarr_get", [storeUrl, key], {
                        signal: AbortSignal.timeout(invokeTimeout),
                    });
                    if (!data.success) return undefined;

                    if (key.includes("spatialdata_attrs") && key.endsWith("0") && !ArrayBuffer.isView(buffers[0].buffer)) {
                        // For some reason, the Zarrita.js UnicodeStringArray does not seem to work with
                        // ArrayBuffers (throws a TypeError), so here we convert to Uint8Array if needed.
                        // This error is occurring specifically for the arr.getChunk call within the AnnDataSource._loadString function.
                        // TODO: figure out a more long-term solution.
                        return new Uint8Array(buffers[0].buffer);
                    }

                    return buffers[0].buffer;
                },
            }
        ])),
    );

    function invokePluginCommand(commandName, commandParams, commandBuffers) {
        return view.experimental.invoke("_plugin_command", [commandName, commandParams], {
            signal: AbortSignal.timeout(invokeTimeout),
            ...(commandBuffers ? { buffers: commandBuffers } : {}),
        });
    }

    for (const pluginEsm of pluginEsmArr) {
        try {
            const pluginEsmUrl = URL.createObjectURL(new Blob([pluginEsm], { type: "text/javascript" }));
            const pluginModule = (await import(pluginEsmUrl)).default;
            URL.revokeObjectURL(pluginEsmUrl);

            const pluginsObj = await pluginModule.createPlugins({
                React,
                PluginFileType,
                PluginViewType,
                PluginCoordinationType,
                PluginJointFileType,
                z,
                invokeCommand: invokePluginCommand,
                useCoordination,
                useGridItemSize,
                useComplexCoordination,
                useMultiCoordinationScopesNonNull,
                useMultiCoordinationScopesSecondaryNonNull,
                useComplexCoordinationSecondary,
                useCoordinationScopes,
                useCoordinationScopesBy,
            });
            if(Array.isArray(pluginsObj.pluginViewTypes)) {
                pluginViewTypes = [...pluginViewTypes, ...pluginsObj.pluginViewTypes];
            }
            if(Array.isArray(pluginsObj.pluginCoordinationTypes)) {
                pluginCoordinationTypes = [...pluginCoordinationTypes, ...pluginsObj.pluginCoordinationTypes];
            }
            if(Array.isArray(pluginsObj.pluginFileTypes)) {
                pluginFileTypes = [...pluginFileTypes, ...pluginsObj.pluginFileTypes];
            }
            if(Array.isArray(pluginsObj.pluginJointFileTypes)) {
                pluginJointFileTypes = [...pluginJointFileTypes, ...pluginsObj.pluginJointFileTypes];
            }
        } catch(e) {
            console.error(e);
        }
    }

    function VitessceWidget(props) {
        const { model } = props;

        const [config, setConfig] = React.useState(prependBaseUrl(model.get('config'), model.get('proxy'), model.get('has_host_name')));
        const [validateConfig, setValidateConfig] = React.useState(true);
        const height = model.get('height');
        const theme = model.get('theme') === 'auto' ? (prefersDark ? 'dark' : 'light') : model.get('theme');

        const divRef = React.useRef();

        React.useEffect(() => {
            if(!divRef.current) {
                return () => {};
            }

            function handleMouseEnter() {
                const jpn = divRef.current.closest('.jp-Notebook');
                if(jpn) {
                    jpn.style.overflow = "hidden";
                }
            }
            function handleMouseLeave(event) {
                if(event.relatedTarget === null || (event.relatedTarget && event.relatedTarget.closest('.jp-Notebook')?.length)) return;
                const jpn = divRef.current.closest('.jp-Notebook');
                if(jpn) {
                    jpn.style.overflow = "auto";
                }
            }
            divRef.current.addEventListener("mouseenter", handleMouseEnter);
            divRef.current.addEventListener("mouseleave", handleMouseLeave);

            return () => {
                if(divRef.current) {
                    divRef.current.removeEventListener("mouseenter", handleMouseEnter);
                    divRef.current.removeEventListener("mouseleave", handleMouseLeave);
                }
            };
        }, [divRef]);

        // Config changed on JS side (from within <Vitessce/>),
        // send updated config to Python side.
        const onConfigChange = React.useCallback((config) => {
            model.set('config', config);
            setValidateConfig(false);
            model.save_changes();
        }, [model]);

        // Config changed on Python side,
        // pass to <Vitessce/> component to it is updated on JS side.
        React.useEffect(() => {
            model.on('change:config', () => {
                const newConfig = prependBaseUrl(model.get('config'), model.get('proxy'), model.get('has_host_name'));

                // Force a re-render and re-validation by setting a new config.uid value.
                // TODO: make this conditional on a parameter from Python.
                //newConfig.uid = `random-${Math.random()}`;
                //console.log('newConfig', newConfig);
                setConfig(newConfig);
            });
        }, []);

        const vitessceProps = {
            height, theme, config, onConfigChange, validateConfig,
            pluginViewTypes, pluginCoordinationTypes, pluginFileTypes, pluginJointFileTypes,
            remountOnUidChange, stores,
        };

        return e('div', { ref: divRef, style: { height: height + 'px' } },
            e(React.Suspense, { fallback: e('div', {}, 'Loading...') },
                e(React.StrictMode, {},
                    e(Vitessce, vitessceProps)
                ),
            ),
        );
    }

    const root = createRoot(view.el);
    root.render(e(VitessceWidget, { model: view.model }));

    return () => {
        // Re-enable scrolling.
        const jpn = view.el.closest('.jp-Notebook');
        if(jpn) {
            jpn.style.overflow = "auto";
        }

        // Clean up React and DOM state.
        root.unmount();
        if(view._isFromDisplay) {
            view.el.remove();
        }
    };
}
export default { render };
"""

DEFAULT_PLUGIN_ESM = """
function createPlugins(utilsForPlugins) {
    const {
        React,
        PluginFileType,
        PluginViewType,
        PluginCoordinationType,
        PluginJointFileType,
        z,
        useCoordination,
        invokeCommand,
    } = utilsForPlugins;
    return {
        pluginViewTypes: undefined,
        pluginFileTypes: undefined,
        pluginCoordinationTypes: undefined,
        pluginJointFileTypes: undefined,
    };
}
export default { createPlugins };
"""


class VitesscePlugin:
    """
    A class that represents a Vitessce widget plugin. Custom plugins can be created by subclassing this class.
    """
    plugin_esm = DEFAULT_PLUGIN_ESM
    commands = {}

    def on_config_change(self, new_config):
        """
        Config change handler.

        :param dict new_config: The new config object.

        :returns: config (likely with new "uid" property) or None
        :rtype: dict or None
        """
        raise NotImplementedError("on_config_change may optionally be implemented by subclasses.")


class VitessceWidget(anywidget.AnyWidget):
    """
    A class to represent a Jupyter widget for Vitessce.
    """
    _esm = ESM

    # Widget specific property.
    # Widget properties are defined as traitlets. Any property tagged with `sync=True`
    # is automatically synced to the frontend *any* time it changes in Python.
    # It is synced back to Python from the frontend *any* time the model is touched.
    config = Dict({}).tag(sync=True)
    height = Int(600).tag(sync=True)
    theme = Unicode('auto').tag(sync=True)
    proxy = Bool(False).tag(sync=True)
    uid = Unicode('').tag(sync=True)
    has_host_name = Bool(False).tag(sync=True)

    next_port = DEFAULT_PORT

    js_package_version = Unicode('3.5.4').tag(sync=True)
    js_dev_mode = Bool(False).tag(sync=True)
    custom_js_url = Unicode('').tag(sync=True)
    plugin_esm = List(trait=Unicode(''), default_value=[]).tag(sync=True)
    remount_on_uid_change = Bool(True).tag(sync=True)
    invoke_timeout = Int(30000).tag(sync=True)

    store_urls = List(trait=Unicode(''), default_value=[]).tag(sync=True)

    def __init__(self, config, height=600, theme='auto', uid=None, port=None, proxy=False, js_package_version='3.5.4', js_dev_mode=False, custom_js_url='', plugins=None, remount_on_uid_change=True, invoke_timeout=30000):
        """
        Construct a new Vitessce widget.

        :param config: A view config instance.
        :type config: VitessceConfig
        :param str theme: The theme name, either "light" or "dark". By default, "auto", which selects light or dark based on operating system preferences.
        :param int height: The height of the widget, in pixels. By default, 600.
        :param int port: The port to use when serving data objects on localhost. By default, 8000.
        :param bool proxy: Is this widget being served through a proxy, for example with a cloud notebook (e.g. Binder)?
        :param str js_package_version: The version of the NPM package ('vitessce' if not js_dev_mode else '@vitessce/dev').
        :param bool js_dev_mode: Should @vitessce/dev be used (typically for debugging purposes)? By default, False.
        :param str custom_js_url: A URL to a JavaScript file to use (instead of 'vitessce' or '@vitessce/dev' NPM package).
        :param list[WidgetPlugin] plugins: A list of subclasses of VitesscePlugin. Optional.
        :param bool remount_on_uid_change: Passed to the remountOnUidChange prop of the <Vitessce/> React component. By default, True.
        :param int invoke_timeout: The timeout in milliseconds for invoking Python functions from JavaScript. By default, 30000.

        .. code-block:: python
            :emphasize-lines: 4

            from vitessce import VitessceConfig, VitessceWidget

            vc = VitessceConfig.from_object(my_scanpy_object)
            vw = vc.widget()
            vw
        """

        base_url, use_port, VitessceWidget.next_port = get_base_url_and_port(
            port, VitessceWidget.next_port, proxy=proxy)
        self.config_obj = config
        self.port = use_port
        config_dict = config.to_dict(base_url=base_url)
        routes = config.get_routes()

        self._stores = config.get_stores(base_url=base_url)
        self._plugins = plugins or []

        plugin_esm = [p.plugin_esm for p in self._plugins]
        self._plugin_commands = {}
        for plugin in self._plugins:
            self._plugin_commands.update(plugin.commands)

        uid_str = get_uid_str(uid)

        super(VitessceWidget, self).__init__(
            config=config_dict, height=height, theme=theme, proxy=proxy,
            js_package_version=js_package_version, js_dev_mode=js_dev_mode, custom_js_url=custom_js_url,
            plugin_esm=plugin_esm, remount_on_uid_change=remount_on_uid_change, invoke_timeout=invoke_timeout,
            uid=uid_str, store_urls=list(self._stores.keys())
        )

        # Register chained plugin on_config_change functions with a change observer.
        def handle_config_change(change):
            new_config = change.new
            for plugin in self._plugins:
                try:
                    new_config = plugin.on_config_change(new_config)
                except NotImplementedError:
                    # It is optional for plugins to implement on_config_change.
                    pass
            if new_config is not None:
                self.config = new_config

        self.observe(handle_config_change, names=['config'])

        serve_routes(config, routes, use_port)

    def _get_coordination_value(self, coordination_type, coordination_scope):
        obj = self.config['coordinationSpace'][coordination_type]
        obj_scopes = list(obj.keys())
        if coordination_scope is not None:
            if coordination_scope in obj_scopes:
                return obj[coordination_scope]
            else:
                raise ValueError(
                    f"The specified coordination scope '{coordination_scope}' could not be found for the coordination type '{coordination_type}'. Known coordination scopes are {obj_scopes}")
        else:
            if len(obj_scopes) == 1:
                auto_coordination_scope = obj_scopes[0]
                return obj[auto_coordination_scope]
            elif len(obj_scopes) > 1:
                raise ValueError(
                    f"The coordination scope could not be automatically determined because multiple coordination scopes exist for the coordination type '{coordination_type}'. Please specify one of {obj_scopes} using the scope parameter.")
            else:
                raise ValueError(
                    f"No coordination scopes were found for the coordination type '{coordination_type}'.")

    def get_cell_selection(self, scope=None):
        return self._get_coordination_value('cellSelection', scope)

    def close(self):
        self.config_obj.stop_server(self.port)
        super().close()

    @anywidget.experimental.command
    def _zarr_get(self, params, buffers):
        [store_url, key] = params
        store = self._stores[store_url]
        try:
            buffers = [store[key.lstrip("/")]]
        except KeyError:
            buffers = []
        return {"success": len(buffers) == 1}, buffers

    @anywidget.experimental.command
    def _plugin_command(self, params, buffers):
        [command_name, command_params] = params
        command_func = self._plugin_commands[command_name]
        return command_func(command_params, buffers)

# Launch Vitessce using plain HTML representation (no ipywidgets)


def ipython_display(config, height=600, theme='auto', base_url=None, host_name=None, uid=None, port=None, proxy=False, js_package_version='3.5.4', js_dev_mode=False, custom_js_url='', plugin_esm=DEFAULT_PLUGIN_ESM, remount_on_uid_change=True):
    from IPython.display import display, HTML
    uid_str = "vitessce" + get_uid_str(uid)

    base_url, use_port, _ = get_base_url_and_port(
        port, DEFAULT_PORT, proxy=proxy, base_url=base_url, host_name=host_name)
    config_dict = config.to_dict(base_url=base_url)
    routes = config.get_routes()
    serve_routes(config, routes, use_port)

    model_vals = {
        "uid": uid_str,
        "js_package_version": js_package_version,
        "js_dev_mode": js_dev_mode,
        "custom_js_url": custom_js_url,
        "plugin_esm": plugin_esm,
        "remount_on_uid_change": remount_on_uid_change,
        "invoke_timeout": 30000,
        "proxy": proxy,
        "has_host_name": host_name is not None,
        "height": height,
        "theme": theme,
        "config": config_dict,
        "store_urls": [],
    }

    # We need to clean up the React and DOM state in any case in which
    # .display() is being run in the same cell as a previous .display().
    # Otherwise, React tries to diff the virtual DOM,
    # causing the browser to become unresponsive.

    # In the .widget() case, we return a cleanup function that anywidget runs for us.
    # However, in the .display() case, we need to run do cleanup ourselves.
    # However, using IPython.display, there is not bidirectional communication,
    # so we cannot simply sent "events" or "messages" to previously rendered widgets
    # to tell them to clean up. Instead, we need to store a reference to the wrapper
    # div element on the window, scoped to the cell, for cleanup.

    # There are two edge cases in which the user runs .display(), then .widget()
    # or vice-versa, in the same cell -- we do not currently handle those,
    # as this would require using the hack-y cleanup over the AnyWidget cleanup
    # in all cases.

    CLEANUP_STR = """
        // Initialize the global Map if not yet defined.
        if (!window.__VITESSCE_DISPLAY_CELLS__) {
            window.__VITESSCE_DISPLAY_CELLS__ = new Map();
        }

        // Rename for readability.
        const CELL_MAP = window.__VITESSCE_DISPLAY_CELLS__;

        // Need cases for getting parent in plain notebook, colab, etc.
        const potentialParents = [
            nextWidgetEl.closest('.jp-Notebook-cell'), // JupyterLab
            nextWidgetEl.closest('.cell'), // JupyterLab Classic Notebook
        ];

        // Get the cell's parent element.
        const parentEl = potentialParents.find(potentialEl => potentialEl !== null);

        if (parentEl) {
            const prevCleanupFunctionPromise = CELL_MAP.get(parentEl);

            if (prevCleanupFunctionPromise) {
                const prevCleanupFunction = await prevCleanupFunctionPromise;
                prevCleanupFunction();
                CELL_MAP.set(parentEl, null);
            }
        }
    """

    HTML_STR = f"""
        <div id="{uid_str}"></div>

        <script type="module">

            const nextWidgetEl = document.getElementById("{uid_str}");

            """ + ESM + """

            """ + CLEANUP_STR + """

            const nextCleanupFunction = render({
                model: {
                    get: (key) => {
                        const vals = """ + json.dumps(model_vals) + """;
                        return vals[key];
                    },
                    set: () => {},
                    save_changes: () => {},
                    on: () => {},
                },
                el: nextWidgetEl,
                _isFromDisplay: true,
            });

            // Store a reference to the widget div element on the window, scoped to the cell,
            // for future cleanups to use.
            if (parentEl) {
                CELL_MAP.set(parentEl, nextCleanupFunction);
            }
        </script>
    """

    display(HTML(HTML_STR))




from starlette.routing import Route
from starlette.responses import StreamingResponse
from pathlib import Path

# Adapted from https://gist.github.com/tombulled/712fd8e19ed0618c5f9f7d5f5f543782


def ranged(file, start=0, end=None, block_size=65535):
    consumed = 0
    file.seek(start)

    while True:
        data_length = (
            min(block_size, end - start - consumed)
            if end else block_size
        )
        if data_length <= 0:
            break
        data = file.read(data_length)
        if not data:
            break
        consumed += data_length
        yield data

    if hasattr(file, 'close'):
        file.close()


def range_repsonse(request, file_path):
    path = Path(file_path)
    file = path.open('rb')
    file_size = path.stat().st_size
    content_range = request.headers.get('range')
    content_length = file_size
    status_code = 200
    headers = {}

    if content_range is not None:
        content_range = content_range.strip().lower()
        content_ranges = content_range.split('=')[-1]
        range_start, range_end, * \
            _ = map(str.strip, (content_ranges + '-').split('-'))
        range_start = max(0, int(range_start)) if range_start else 0
        range_end = min(file_size - 1, int(range_end)
                        ) if range_end else file_size - 1
        content_length = (range_end - range_start) + 1
        file = ranged(file, start=range_start, end=range_end + 1)
        status_code = 206
        headers['Content-Range'] = f'bytes {range_start}-{range_end}/{file_size}'

    response = StreamingResponse(
        file,
        media_type='tiff',
        status_code=status_code,
    )
    response.headers.update({
        'Accept-Ranges': 'bytes',
        'Content-Length': str(content_length),
        **headers,
    })
    return response


class JsonRoute(Route):
    def __init__(self, path, endpoint, data_json):
        super().__init__(path, endpoint)
        self.data_json = data_json


class FileRoute(Route):
    def __init__(self, path, endpoint, file_path):
        super().__init__(path, endpoint)
        self.file_path = file_path



import numpy as np
import zarr
from ome_zarr.writer import write_image
from tifffile import TiffWriter
from .anndata import cast_arr


def needs_bigtiff(img_arr_shape):
    """
    Helper function to determine if an image array is too large for standard TIFF format.

    :param img_arr_shape: The shape of the image array.
    :type img_arr_shape: tuple[int]
    :return: True if the image array is too large for standard TIFF format, False otherwise.
    :rtype: bool
    """
    num_pixels = 1
    for n in img_arr_shape.shape:
        num_pixels *= n
    return (num_pixels > 2**32)


def rgb_img_to_ome_tiff(img_arr, output_path, img_name="Image", axes="CYX"):
    """
    Convert an RGB image to OME-TIFF.

    :param img_arr: The image as a 3D array.
    :type img_arr: np.array
    :param str output_path: The path to save the Zarr store.
    :param str img_name: The name of the image to include in the omero.name NGFF metadata field.
    :param str axes: The array axis ordering. By default, "CYX"
    """
    img_arr = img_arr.astype(np.dtype('uint8'))
    bigtiff = needs_bigtiff(img_arr.shape)

    tiff_writer = TiffWriter(output_path, ome=True, bigtiff=bigtiff)
    tiff_writer.write(
        img_arr,
        metadata={
            'axes': axes,
            'Channel': {'Name': ['R', 'G', 'B']},
        }
    )
    tiff_writer.close()


def multiplex_img_to_ome_tiff(img_arr, channel_names, output_path, axes="CYX"):
    """
    Convert a multiplexed image to OME-TIFF.

    :param img_arr: The image as a 3D, 4D, or 5D array.
    :type img_arr: np.array
    :param list[str] channel_names: A list of channel names to include in the omero.channels[].label NGFF metadata field.
    :param str output_path: The path to save the Zarr store.
    :param str axes: The array axis ordering. By default, "CYX"
    """
    bigtiff = needs_bigtiff(img_arr.shape)

    tiff_writer = TiffWriter(output_path, ome=True, bigtiff=bigtiff)
    tiff_writer.write(
        img_arr,
        metadata={
            'axes': axes,
            'Channel': {'Name': channel_names},
        }
    )
    tiff_writer.close()


def rgb_img_to_ome_zarr(img_arr, output_path, img_name="Image", chunks=(1, 256, 256), axes="cyx", **kwargs):
    """
    Convert an RGB image to OME-Zarr v0.3.

    :param img_arr: The image as a 3D array.
    :type img_arr: np.array
    :param str output_path: The path to save the Zarr store.
    :param str img_name: The name of the image to include in the omero.name NGFF metadata field.
    :param tuple[int] chunks: The chunk sizes of each axis. By default, (1, 256, 256).
    :param str axes: The array axis ordering. By default, "cyx"
    """
    img_arr = img_arr.astype(np.dtype('uint8'))

    default_window = {
        "start": 0,
        "min": 0,
        "max": 255,
        "end": 255
    }

    z_root = zarr.open_group(output_path, mode="w")

    write_image(
        image=img_arr,
        group=z_root,
        axes=axes,
        storage_options=dict(chunks=chunks),
        **kwargs,
    )
    z_root.attrs["omero"] = {
        "name": img_name,
        "version": "0.3",
        "rdefs": {},
        "channels": [
            {
                "label": "R",
                "color": "FF0000",
                "window": default_window
            },
            {
                "label": "G",
                "color": "00FF00",
                "window": default_window
            },
            {
                "label": "B",
                "color": "0000FF",
                "window": default_window
            }
        ]
    }


def multiplex_img_to_ome_zarr(img_arr, channel_names, output_path, img_name="Image", chunks=(1, 256, 256), axes="cyx", channel_colors=None):
    """
    Convert a multiplexed image to OME-Zarr v0.3.

    :param img_arr: The image as a 3D, 4D, or 5D array.
    :type img_arr: np.array
    :param list[str] channel_names: A list of channel names to include in the omero.channels[].label NGFF metadata field.
    :param str output_path: The path to save the Zarr store.
    :param str img_name: The name of the image to include in the omero.name NGFF metadata field.
    :param tuple[int] chunks: The chunk sizes of each axis. By default, (1, 256, 256).
    :param str axes: The array axis ordering. By default, "cyx"
    :param channel_colors: Dict mapping channel names to color strings to use for the omero.channels[].color NGFF metadata field. If provided, keys should match channel_names. By default, None to use "FFFFFF" for all channels.
    :type channel_colors: dict or None
    """
    img_arr = cast_arr(img_arr)

    dtype_info = np.iinfo(img_arr.dtype) if img_arr.dtype.kind == 'u' or img_arr.dtype.kind == 'i' else np.finfo(img_arr.dtype)

    default_window = {
        "start": 0,
        "min": 0,
        "max": dtype_info.max,
        "end": dtype_info.max
    }

    z_root = zarr.open_group(output_path, mode="w")

    write_image(
        image=img_arr,
        group=z_root,
        axes=axes,
        storage_options=dict(chunks=chunks)
    )
    z_root.attrs["omero"] = {
        "name": img_name,
        "version": "0.3",
        "rdefs": {},
        "channels": [
            {
                "label": channel_name,
                "color": channel_colors[channel_name] if channel_colors is not None else "FFFFFF",
                "window": default_window
            }
            for channel_name
            in channel_names
        ]
    }



from enum import Enum


BASE_URL_PLACEHOLDER = "{{ base_url }}"


# Reference: https://stackoverflow.com/a/50473952
class DocEnum(Enum):
    def __new__(cls, value, doc):
        self = object.__new__(cls)
        self._value_ = value
        self.__doc__ = doc
        return self


def norm_enum(enum_val, expected_enum_class=None):
    # assert isinstance(enum_val, str) or isinstance(enum_val, expected_enum_class), f"enum_val was {type(enum_val)} and not a string or expected value {type(expected_enum_class)}"
    # We don't actually use the expected_enum_class,
    # since it would not account for things like plugin coordination types, etc.
    # But we can pass it around anyway and in the future could use
    # it for extra checks, e.g. upon debug flags being true.
    if isinstance(enum_val, str):
        return enum_val
    else:
        return enum_val.value


class CoordinationType(DocEnum):
    """
    An enum type representing a coordination type in the Vitessce coordination model.
    The term coordination type refers to a parameter to be coordinated, and its programming-language-like type.
    For example, the ``SPATIAL_ZOOM`` coordination type represents a coordination of the zoom level of a spatial view, which can take a float value.
    """
    META_COORDINATION_SCOPES = "metaCoordinationScopes", "Shared representation of view-level coordinationScopes."
    META_COORDINATION_SCOPES_BY = "metaCoordinationScopesBy", "Shared representation of view-level coordinationScopesBy."
    DATASET = "dataset", "The identifier for the dataset associated with a view."
    OBS_TYPE = "obsType", "The type of entity represented by each observation."
    FEATURE_TYPE = "featureType", "The type of entity represented by each feature."
    FEATURE_VALUE_TYPE = "featureValueType", "The type of value stored in the observation-by-feature matrix."
    OBS_LABELS_TYPE = 'obsLabelsType', "Feature for displaying additional obs sets' data in heatmap/scatterplot/spatial tooltips."
    EMBEDDING_TYPE = "embeddingType", "The type of embedding used for a scatterplot view, such as PCA or t-SNE."
    EMBEDDING_ZOOM = "embeddingZoom", "The zoom level of an embedding scatterplot view."
    EMBEDDING_ROTATION = "embeddingRotation", "The rotation of an embedding scatterplot view."
    EMBEDDING_TARGET_X = "embeddingTargetX", "The x-coordinate of the center of an embedding scatterplot view."
    EMBEDDING_TARGET_Y = "embeddingTargetY", "The y-coordinate of the center of an embedding scatterplot view."
    EMBEDDING_TARGET_Z = "embeddingTargetZ", "The z-coordinate of the center of an embedding scatterplot view."
    EMBEDDING_OBS_SET_POLYGONS_VISIBLE = 'embeddingObsSetPolygonsVisible', "Whether polygon boundaries for each selected obsSet are visible in the embedding scatterplt."
    EMBEDDING_OBS_SET_LABELS_VISIBLE = 'embeddingObsSetLabelsVisible', "Whether labels for each selected obsSet are visible in the embedding scatterplot."
    EMBEDDING_OBS_SET_LABEL_SIZE = 'embeddingObsSetLabelSize', "The size of labels for selected obsSets in the embedding scatterplot."
    EMBEDDING_OBS_OPACITY = 'embeddingObsOpacity', "Opacity of cells in embedding (points in scatterplot view)"
    EMBEDDING_OBS_RADIUS = 'embeddingObsRadius', "Radius of cells in embedding (points in scatterplot view)"
    EMBEDDING_OBS_RADIUS_MODE = 'embeddingObsRadiusMode', "Radius mode of cells in embedding (points in scatterplot view) - auto or manual"
    EMBEDDING_OBS_OPACITY_MODE = 'embeddingObsOpacityMode', "Opacity mode of cells in embedding (points in scatterplot view) - auto or manual"
    EMBEDDING_CELL_OPACITY = 'embeddingCellOpacity', "Deprecated"
    EMBEDDING_CELL_RADIUS = 'embeddingCellRadius', "Deprecated"
    EMBEDDING_CELL_RADIUS_MODE = 'embeddingCellRadiusMode', "Deprecated"
    EMBEDDING_CELL_OPACITY_MODE = 'embeddingCellOpacityMode', "Deprecated"
    SPATIAL_ZOOM = "spatialZoom", "The zoom level of a spatial view."
    SPATIAL_ROTATION_X = 'spatialRotationX', "The x rotation of a 3d spatial view."
    SPATIAL_ROTATION_Y = 'spatialRotationY', "The y rotation of a 3d spatial view."
    SPATIAL_ROTATION_Z = 'spatialRotationZ', "The z rotation of a 3d spatial view."
    SPATIAL_ROTATION_ORBIT = 'spatialRotationOrbit', "The rotation orbit in degrees of a 3d spatial view."
    SPATIAL_ORBIT_AXIS = 'spatialOrbitAxis', "The orbital axis of a 3d spatial view."
    SPATIAL_AXIS_FIXED = 'spatialAxisFixed', "Boolean for whether or not the target axis of a spatial view is fixed."
    SPATIAL_TARGET_X = "spatialTargetX", "The x-coordinate of the center of a spatial view."
    SPATIAL_TARGET_Y = "spatialTargetY", "The y-coordinate of the center of a spatial view."
    SPATIAL_TARGET_Z = "spatialTargetZ", "The z-coordinate of the center of a spatial view."
    HEATMAP_ZOOM_X = "heatmapZoomX", "The x-axis zoom level of a heatmap view."
    HEATMAP_ZOOM_Y = "heatmapZoomY", "The y-axis zoom level of a heatmap view."
    HEATMAP_TARGET_X = "heatmapTargetX", "The x-coordinate of the center of a heatmap view."
    HEATMAP_TARGET_Y = "heatmapTargetY", "The y-coordinate of the center of a heatmap view."
    OBS_FILTER = "obsFilter", "A subset of cells to include after filtering."
    OBS_HIGHLIGHT = "obsHighlight", "A subset of cells to highlight."
    OBS_SET_SELECTION = "obsSetSelection", "A subset of cell sets to select."
    OBS_SET_HIGHLIGHT = "obsSetHighlight", "A subset of cell sets to highlight."
    OBS_SET_COLOR = "obsSetColor", "A mapping from cell sets to colors."
    CELL_FILTER = "cellFilter", "Deprecated"
    CELL_HIGHLIGHT = "cellHighlight", "Deprecated"
    CELL_SET_SELECTION = "cellSetSelection", "Deprecated"
    CELL_SET_HIGHLIGHT = "cellSetHighlight", "Deprecated"
    CELL_SET_COLOR = "cellSetColor", "Deprecated"
    FEATURE_FILTER = "featureFilter", "A subset of genes to include after filtering."
    FEATURE_HIGHLIGHT = "featureHighlight", "A subset of genes to highlight."
    FEATURE_SELECTION = "featureSelection", "A subset of genes to select."
    FEATURE_VALUE_COLORMAP = "featureValueColormap", "The colormap to use for the gene expression scale."
    FEATURE_VALUE_TRANSFORM = 'featureValueTransform', "Function to use to transform feature values."
    FEATURE_VALUE_COLORMAP_RANGE = "featureValueColormapRange", "The range of gene expression values to map."
    GENE_FILTER = "geneFilter", "Deprecated"
    GENE_HIGHLIGHT = "geneHighlight", "Deprecated"
    GENE_SELECTION = "geneSelection", "Deprecated"
    GENE_EXPRESSION_COLORMAP = "geneExpressionColormap", "Deprecated"
    GENE_EXPRESSION_COLORMAP_RANGE = "geneExpressionColormapRange", "Deprecated"
    OBS_COLOR_ENCODING = "obsColorEncoding", "The color encoding to use for cell entities."
    CELL_COLOR_ENCODING = "cellColorEncoding", "Deprecated"
    SPATIAL_IMAGE_LAYER = 'spatialImageLayer', "Layer definitions for the imagery in the spatial view."
    SPATIAL_SEGMENTATION_LAYER = 'spatialSegmentationLayer', "Layer definitions for the segmentations in the spatial view."
    SPATIAL_POINT_LAYER = 'spatialPointLayer', "Layer definitions for the points in the spatial view."
    SPATIAL_NEIGHBORHOOD_LAYER = 'spatialNeighborhoodLayer', "Layer definitions for the neighborhoods in the spatial view."
    SPATIAL_RASTER_LAYERS = 'spatialRasterLayers', "Deprecated"
    SPATIAL_CELLS_LAYER = 'spatialCellsLayer', "Deprecated"
    SPATIAL_MOLECULES_LAYER = 'spatialMoleculesLayer', "Deprecated"
    SPATIAL_NEIGHBORHOODS_LAYER = 'spatialNeighborhoodsLayer', "Deprecated"
    GENOMIC_ZOOM_X = "genomicZoomX", "The zoom level of a higlass view, X dimension."
    GENOMIC_ZOOM_Y = "genomicZoomY", "The zoom level of a higlass view, Y dimension."
    GENOMIC_TARGET_X = "genomicTargetX", "The x-coordinate of the center of a higlass view."
    GENOMIC_TARGET_Y = "genomicTargetY", "The y-coordinate of the center of a higlass view."
    ADDITIONAL_CELL_SETS = "additionalCellSets", "Deprecated"
    ADDITIONAL_OBS_SETS = "additionalObsSets", "User-defined cell sets."
    MOLECULE_HIGHLIGHT = 'moleculeHighlight', "Deprecated"
    GATING_FEATURE_SELECTION_X = 'gatingFeatureSelectionX', "Feature for the x-axis of the gating scatterplot."
    GATING_FEATURE_SELECTION_Y = 'gatingFeatureSelectionY', "Feature for the y-axis of the gating scatterplot."
    FEATURE_VALUE_TRANSFORM_COEFFICIENT = 'featureValueTransformCoefficient', "Coefficient to transform values in the gating scatterplot."
    TOOLTIPS_VISIBLE = 'tooltipsVisible', "Boolean for whether or not tooltips are visible, used by the scatterplot, spatial, and heatmap views."


class ViewType(DocEnum):
    """
    An enum type representing a view type in the visualization layout.
    """
    SCATTERPLOT = "scatterplot", "The scatterplot component can be used for visualization of 2-dimensional embeddings."
    SPATIAL = "spatial", "The spatial component can be used for visualization of cells, molecules, or images in spatial coordinates."
    DESCRIPTION = "description", "The description component can display short informational text about a dataset."
    STATUS = "status", "The status component can display contextual information such as hover states or error messages."
    HEATMAP = "heatmap", "The heatmap component can be used to view a cell by gene expression matrix."
    LAYER_CONTROLLER = "layerController", "The layer controller can be used to manipulate channel settings of the images rendered by the spatial component."
    GENOMIC_PROFILES = "genomicProfiles", "The higlass component can be used to visualize genome-wide ATAC-seq profiles."
    OBS_SETS = "obsSets", "Observation sets"
    OBS_SET_SIZES = "obsSetSizes", "Observation set sizes bar plot"
    OBS_SET_FEATURE_VALUE_DISTRIBUTION = "obsSetFeatureValueDistribution", "Violin plot visualizing the distribution of feature values per observation set."
    FEATURE_VALUE_HISTOGRAM = "featureValueHistogram", "Histogram visualizing the distribution of values for a selected feature across all observations."
    FEATURE_LIST = "featureList", "The feature list selector"
    GATING = "gating", "A gating scatterplot"
    CELL_SETS = "cellSets", "Deprecated"
    CELL_SET_SIZES = "cellSetSizes", "Deprecated"
    CELL_SET_EXPRESSION = "cellSetExpression", "Deprecated"
    GENES = "genes", "Deprecated"


class DataType(DocEnum):
    """
    An enum type representing the type of data contained in a file.
    """
    OBS_LABELS = "obsLabels", "Alternate label for each observation."
    OBS_EMBEDDING = "obsEmbedding", "Embedding coordinates for each observation."
    OBS_LOCATIONS = "obsLocations", "Spatial coordinates for each observation."
    OBS_FEATURE_MATRIX = "obsFeatureMatrix", "Matrix of feature (column) values for each observation (row)."
    OBS_SETS = "obsSets", "Sets of observations, representing clusters or classes."
    FEATURE_LABELS = "featureLabels", "Alternate label for each feature."
    IMAGE = "image", "Image"
    OBS_SEGMENTATIONS = "obsSegmentations", "Segmentation for each observation."
    NEIGHBORHOODS = "neighborhoods", "The spatial cell neighborhoods data type."
    GENOMIC_PROFILES = "genomic-profiles", "The genomic profiles data type, used by HiGlass 1D quantitative tracks."
    CELLS = "cells", "Deprecated"
    CELL_SETS = "cell-sets", "Deprecated"
    EXPRESSION_MATRIX = "expression-matrix", "Deprecated"
    MOLECULES = "molecules", "Deprecated"
    RASTER = "raster", "Deprecated"


class FileType(DocEnum):
    """
    An enum type representing the file format or schema to which a file conforms.
    """
    ANNDATA_ZARR = "anndata.zarr", "Joint file type for AnnData objects"
    SPATIALDATA_ZARR = "spatialdata.zarr", "Joint file type for SpatialData objects"
    ANNDATA_H5AD = "anndata.h5ad", "Joint file type for AnnData objects"
    OBS_EMBEDDING_CSV = 'obsEmbedding.csv', "File type for obsEmbedding values stored in a CSV file"
    OBS_LOCATIONS_CSV = 'obsLocations.csv', "File type for obsLocations values stored in a CSV file"
    OBS_LABELS_CSV = 'obsLabels.csv', "File type for obsLabels values stored in a CSV file"
    FEATURE_LABELS_CSV = 'featureLabels.csv', "File type for featureLabels values stored in a CSV file"
    OBS_FEATURE_MATRIX_CSV = 'obsFeatureMatrix.csv', "File type for obsFeatureMatrix stored in a CSV file"
    OBS_SEGMENTATIONS_JSON = 'obsSegmentations.json', "File type for obsSegmentations polygons stored in a JSON file"
    OBS_SETS_CSV = 'obsSets.csv', "File type for obsSets stored in a CSV file"
    OBS_SETS_JSON = 'obsSets.json', "File type for obsSets stored in a JSON file"
    IMAGE_OME_ZARR = "image.ome-zarr", "File type for images stored as OME-NGFF Zarr stores."
    OBS_FEATURE_MATRIX_ANNDATA_ZARR = 'obsFeatureMatrix.anndata.zarr', "File type for obsFeatureMatrix stored in an AnnData object saved to a Zarr store"
    OBS_SETS_ANNDATA_ZARR = 'obsSets.anndata.zarr', "File type for obsSets stored in an AnnData object saved to a Zarr store"
    OBS_EMBEDDING_ANNDATA_ZARR = 'obsEmbedding.anndata.zarr', "File type for obsEmbedding values stored in an AnnData object saved to a Zarr store"
    OBS_LOCATIONS_ANNDATA_ZARR = 'obsLocations.anndata.zarr', "File type for obsLocations values stored in an AnnData object saved to a Zarr store"
    OBS_SEGMENTATIONS_ANNDATA_ZARR = 'obsSegmentations.anndata.zarr', "File type for obsSegmentations polygons stored in an AnnData object saved to a Zarr store"
    OBS_LABELS_ANNDATA_ZARR = 'obsLabels.anndata.zarr', "File type for obsLabels stored in an AnnData object saved to a Zarr store"
    FEATURE_LABELS_ANNDATA_ZARR = 'featureLabels.anndata.zarr', "File type for featureLabels stored in an AnnData object saved to a Zarr store"
    EXPRESSION_MATRIX_ZARR = "expression-matrix.zarr", "The Zarr-based expression matrix file type."
    CELLS_JSON = "cells.json", "The JSON-based cells file type."
    MOLECULES_JSON = "molecules.json", "The JSON-based molecules file type."
    NEIGHBORHOODS_JSON = "neighborhoods.json", "The JSON-based neighborhoods file type."
    RASTER_JSON = "raster.json", "The JSON-based raster manifest file type."
    CELL_SETS_JSON = "cell-sets.json", "The JSON-based cell sets file type."
    CLUSTERS_JSON = "clusters.json", "A JSON-based expression matrix file type (this file type is poorly named)."
    GENES_JSON = "genes.json", "A JSON-based expression matrix file type."
    GENOMIC_PROFILES_ZARR = "genomic-profiles.zarr", "The Zarr-based genomic profile (multivec) file type."
    ANNDATA_CELLS_ZARR = "anndata-cells.zarr", "The Zarr-based cells file type from an anndata object."
    ANNDATA_CELL_SETS_ZARR = "anndata-cell-sets.zarr", "The Zarr-based cell-sets file type from an anndata object."
    ANNDATA_EXPRESSION_MATRIX_ZARR = "anndata-expression-matrix.zarr", "The Zarr-based expression matrix file type from an anndata object."
    OBS_SEGMENTATIONS_CELLS_JSON = "obsSegmentations.cells.json", "The JSON-based cells file type for obsSegmentations."
    OBS_LOCATIONS_CELLS_JSON = "obsLocations.cells.json", "The JSON-based cells file type for obsLocations."
    OBS_EMBEDDING_CELLS_JSON = "obsEmbedding.cells.json", "The JSON-based cells file type for obsEmbedding."
    OBS_SETS_CELL_SETS_JSON = "obsSets.cell-sets.json", "The JSON-based cell sets file type for obsSets."



from __future__ import annotations

from functools import partial
from typing import Optional

import numpy as np


def gen_obs_embedding_schema(options: dict, paths: Optional[list[str]] = None, names: Optional[list[str]] = None, dims: Optional[list[list[int]]] = None):
    if paths is not None:
        if "obsEmbedding" not in options:
            options["obsEmbedding"] = []
        if names is not None:
            for key, mapping in zip(paths, names):
                options["obsEmbedding"].append({
                    "path": key,
                    "dims": [0, 1],
                    "embeddingType": mapping
                })
        else:
            for mapping in paths:
                mapping_key = mapping.split('/')[-1]
                options["obsEmbedding"].append({
                    "path": mapping,
                    "dims": [0, 1],
                    "embeddingType": mapping_key
                })
    if dims is not None:
        if "obsEmbedding" not in options:
            options["obsEmbedding"] = []
        for dim_i, dim in enumerate(dims):
            options["obsEmbedding"][dim_i]['dims'] = dim
    return options


def gen_obs_sets_schema(options: dict, paths: Optional[list[str]] = None, names: Optional[list[str]] = None):
    if paths is not None:
        options["obsSets"] = []
        if names is not None:
            names = names
        else:
            names = []
            for obs in paths:
                obs_end_path = obs.split('/')[-1]
                names += [obs_end_path]
        for obs, name in zip(paths, names):
            options["obsSets"].append({
                "name": name,
                "path": obs
            })
    return options


def gen_sdata_obs_sets_schema(options: dict, paths: Optional[list[str]] = None, names: Optional[list[str]] = None, table_path: Optional[str] = None, region: Optional[str] = None):
    if paths is not None:
        options["obsSets"] = {"obsSets": []}
        if names is not None:
            names = names
        else:
            names = []
            for obs in paths:
                obs_end_path = obs.split('/')[-1]
                names += [obs_end_path]
        for obs, name in zip(paths, names):
            options["obsSets"]["obsSets"].append({
                "name": name,
                "path": obs
            })
        if table_path is not None:
            options["obsSets"]["tablePath"] = table_path
        if region is not None:
            options["obsSets"]["region"] = region
    return options


def gen_obs_feature_matrix_schema(options: dict, matrix_path: Optional[str] = None, var_filter_path: Optional[str] = None, init_var_filter_path: Optional[str] = None):
    if matrix_path is not None:
        options["obsFeatureMatrix"] = {
            "path": matrix_path
        }
        if var_filter_path is not None:
            options["obsFeatureMatrix"]["featureFilterPath"] = var_filter_path
        if init_var_filter_path is not None:
            options["obsFeatureMatrix"]["initialFeatureFilterPath"] = init_var_filter_path
    return options


def gen_obs_labels_schema(options: dict, paths: Optional[list[str]] = None, names: Optional[list[str]] = None):
    if paths is not None:
        if names is not None and len(paths) == len(names):
            # A name was provided for each path element, so use those values.
            names = names
        else:
            # Names were not provided for each path element,
            # so fall back to using the final part of each path for the names.
            names = [labels_path.split('/')[-1] for labels_path in paths]
        obs_labels = []
        for path, name in zip(paths, names):
            obs_labels.append({"path": path, "obsLabelsType": name})
        options["obsLabels"] = obs_labels
    return options


def gen_path_schema(key: str, path: Optional[str], options: dict):
    if path is not None:
        options[key] = {
            "path": path
        }
    return options


gen_obs_locations_schema = partial(gen_path_schema, "obsLocations")
gen_obs_segmentations_schema = partial(gen_path_schema, "obsSegmentations")
gen_obs_spots_schema = partial(gen_path_schema, "obsSpots")
gen_obs_points_schema = partial(gen_path_schema, "obsPoints")
gen_feature_labels_schema = partial(gen_path_schema, "featureLabels")


def gen_sdata_image_schema(options, path: str, coordinate_system: Optional[str] = None, affine_transformation: Optional[np.ndarray] = None) -> dict:
    if path is not None:
        options["image"] = {
            "path": path
        }
        if affine_transformation is not None:
            options["image"]['coordinateTransformations'] = affine_transformation
        if coordinate_system is not None:
            options["image"]['coordinateSystem'] = coordinate_system
    return options


def gen_sdata_labels_schema(options, path: str, table_path: str = "tables/table", coordinate_system: Optional[str] = None, affine_transformation: Optional[np.ndarray] = None) -> dict:
    if path is not None:
        options["labels"] = {
            "path": path
        }
        if table_path is not None:
            options["labels"]['tablePath'] = table_path
        if affine_transformation is not None:
            options["labels"]['coordinateTransformations'] = affine_transformation
        if coordinate_system is not None:
            options["labels"]['coordinateSystem'] = coordinate_system
    return options


def gen_sdata_obs_spots_schema(options: dict, shapes_path: str, table_path: str = "tables/table", region: Optional[str] = None, coordinate_system: Optional[str] = None) -> dict:
    if shapes_path is not None:
        options['obsSpots'] = {
            "path": shapes_path,
            "tablePath": table_path
        }
        if region is not None:
            options['obsSpots']['region'] = region
        if coordinate_system is not None:
            options['obsSpots']['coordinateSystem'] = coordinate_system
    return options


def gen_sdata_obs_feature_matrix_schema(options: dict, matrix_path: Optional[str] = None, var_filter_path: Optional[str] = None, init_var_filter_path: Optional[str] = None, region: Optional[str] = None):
    if matrix_path is not None:
        options["obsFeatureMatrix"] = {
            "path": matrix_path
        }
        if region is not None:
            options['obsFeatureMatrix']['region'] = region
        if var_filter_path is not None:
            options["obsFeatureMatrix"]["featureFilterPath"] = var_filter_path
        if init_var_filter_path is not None:
            options["obsFeatureMatrix"]["initialFeatureFilterPath"] = init_var_filter_path
    return options



import unittest

from pathlib import Path, PurePosixPath, PureWindowsPath

from .create_test_data import (
    create_test_anndata_file,
    create_test_loom_file,
    create_test_ometiff_file,
    create_test_omezarr_store,
    create_test_snaptools_files,
)

from vitessce import (
    OmeTiffWrapper,
    OmeZarrWrapper,
    AnnDataWrapper,
    CsvWrapper,
    JsonWrapper,
    MultivecZarrWrapper,
    ImageOmeTiffWrapper,
    ObsSegmentationsOmeTiffWrapper,
    ImageOmeZarrWrapper,
    ObsSegmentationsOmeZarrWrapper,
)

from vitessce.wrappers import SpatialDataWrapper, file_path_to_url_path


data_path = Path('tests/data')


class TestWrappers(unittest.TestCase):

    def setUp(self):
        create_test_anndata_file(data_path / 'test.h5ad')
        create_test_loom_file(data_path / 'test.loom')
        create_test_ometiff_file(data_path / 'test.ome.tif')
        create_test_omezarr_store(data_path / 'test.ome.zarr')
        create_test_snaptools_files(
            data_path / 'test.snap.mtx',
            data_path / 'test.snap.bins.txt',
            data_path / 'test.snap.barcodes.txt',
            data_path / 'test.snap.clusters.csv',
        )

    def test_file_path_to_url_path(self):
        posix_with_slash = file_path_to_url_path("tests/data/test.snap.mtx", path_class=PurePosixPath)
        self.assertEqual(posix_with_slash, "/tests/data/test.snap.mtx")

        posix_without_slash = file_path_to_url_path("tests/data/test.snap.mtx", prepend_slash=False, path_class=PurePosixPath)
        self.assertEqual(posix_without_slash, "tests/data/test.snap.mtx")

        posix_with_dot_and_slash = file_path_to_url_path("./tests/data/test.snap.mtx", path_class=PurePosixPath)
        self.assertEqual(posix_with_dot_and_slash, "/tests/data/test.snap.mtx")

        posix_with_dot_without_slash = file_path_to_url_path("./tests/data/test.snap.mtx", prepend_slash=False, path_class=PurePosixPath)
        self.assertEqual(posix_with_dot_without_slash, "tests/data/test.snap.mtx")

        windows_with_slash = file_path_to_url_path("tests\\data\\test.snap.mtx", path_class=PureWindowsPath)
        self.assertEqual(windows_with_slash, "/tests/data/test.snap.mtx")

    def test_ome_tiff(self):
        w = OmeTiffWrapper(img_path=data_path / 'test.ome.tif', name="Test")
        w.local_img_uid = 'test.ome.tif'
        w.local_offsets_uid = 'test.offsets.json'

        file_def_creator = w.make_raster_file_def_creator(
            "A",
            "0"
        )
        file_def = file_def_creator('http://localhost:8000')
        self.assertEqual(file_def, {
            'fileType': 'raster.json',
            'options': {
                'schemaVersion': '0.0.2',
                'images': [
                    {
                        'name': 'Test',
                        'type': 'ome-tiff',
                        'url': 'http://localhost:8000/A/0/test.ome.tif',
                        'metadata': {
                            'isBitmask': False,
                            'omeTiffOffsetsUrl': 'http://localhost:8000/A/0/test.offsets.json'
                        }
                    }
                ],
            }
        })

    def test_ome_tiff_with_base_dir(self):
        w = OmeTiffWrapper(img_path='test.ome.tif', name="Test")
        w.base_dir = str(data_path)
        w.local_img_uid = 'test.ome.tif'
        w.local_offsets_uid = 'test.offsets.json'

        file_def_creator = w.make_raster_file_def_creator(
            "A",
            "0"
        )
        file_def = file_def_creator('http://localhost:8000')
        self.assertEqual(file_def, {
            'fileType': 'raster.json',
            'options': {
                'schemaVersion': '0.0.2',
                'images': [
                    {
                        'name': 'Test',
                        'type': 'ome-tiff',
                        'url': 'http://localhost:8000/test.ome.tif',
                        'metadata': {
                            'isBitmask': False,
                        }
                    }
                ],
            }
        })

    def test_ome_zarr(self):
        w = OmeZarrWrapper(img_path=data_path / 'test.ome.zarr')
        w.local_dir_uid = 'test.ome.zarr'

        file_def_creator = w.make_image_file_def_creator(
            "A",
            "0"
        )
        file_def = file_def_creator('http://localhost:8000')
        self.assertEqual(file_def, {
            'fileType': 'image.ome-zarr',
            'url': 'http://localhost:8000/A/0/test.ome.zarr'
        })

    def test_ome_zarr_with_base_dir(self):
        w = OmeZarrWrapper(img_path='test.ome.zarr')
        w.base_dir = str(data_path)
        w.local_dir_uid = 'test.ome.zarr'

        file_def_creator = w.make_image_file_def_creator(
            "A",
            "0"
        )
        file_def = file_def_creator('http://localhost:8000')
        self.assertEqual(file_def, {
            'fileType': 'image.ome-zarr',
            'url': 'http://localhost:8000/test.ome.zarr'
        })

    def test_image_ome_tiff(self):
        w = ImageOmeTiffWrapper(img_path=data_path / 'test.ome.tif')
        w.local_img_uid = 'test.ome.tif'
        w.local_offsets_uid = 'test.offsets.json'

        file_def_creator = w.make_raster_file_def_creator(
            "A",
            "0"
        )
        file_def = file_def_creator('http://localhost:8000')
        self.assertEqual(file_def, {
            'fileType': 'image.ome-tiff',
            'url': 'http://localhost:8000/A/0/test.ome.tif',
            'options': {
                'offsetsUrl': 'http://localhost:8000/A/0/test.offsets.json'
            }
        })

    def test_image_ome_zarr(self):
        w = ImageOmeZarrWrapper(img_path=data_path / 'test.ome.zarr')
        w.local_dir_uid = 'test.ome.zarr'

        file_def_creator = w.make_image_file_def_creator(
            "A",
            "0"
        )
        file_def = file_def_creator('http://localhost:8000')
        self.assertEqual(file_def, {
            'fileType': 'image.ome-zarr',
            'url': 'http://localhost:8000/A/0/test.ome.zarr'
        })

    def test_obs_segmentations_ome_tiff(self):
        w = ObsSegmentationsOmeTiffWrapper(img_path=data_path / 'test.ome.tif')
        w.local_img_uid = 'test.ome.tif'
        w.local_offsets_uid = 'test.offsets.json'

        file_def_creator = w.make_raster_file_def_creator(
            "A",
            "0"
        )
        file_def = file_def_creator('http://localhost:8000')
        self.assertEqual(file_def, {
            'fileType': 'obsSegmentations.ome-tiff',
            'url': 'http://localhost:8000/A/0/test.ome.tif',
            'options': {
                'offsetsUrl': 'http://localhost:8000/A/0/test.offsets.json'
            }
        })

    def test_obs_segmentations_ome_zarr(self):
        w = ObsSegmentationsOmeZarrWrapper(img_path=data_path / 'test.ome.zarr')
        w.local_dir_uid = 'test.ome.zarr'

        file_def_creator = w.make_image_file_def_creator(
            "A",
            "0"
        )
        file_def = file_def_creator('http://localhost:8000')
        self.assertEqual(file_def, {
            'fileType': 'obsSegmentations.ome-zarr',
            'url': 'http://localhost:8000/A/0/test.ome.zarr'
        })

    def test_anndata(self):
        adata_path = data_path / 'test.h5ad.zarr'
        w = AnnDataWrapper(adata_path,
                           obs_set_paths=['obs/CellType'], obs_set_names=['Cell Type'],
                           obs_labels_names=['Cell Label'], obs_labels_paths=['obs/CellLabel'],
                           obs_embedding_paths=['obsm/X_umap'], obs_embedding_names=['UMAP'])
        w.local_dir_uid = 'anndata.zarr'

        file_def_creator = w.make_file_def_creator('A', 0)
        file_def = file_def_creator('http://localhost:8000')
        self.assertEqual(file_def, {'fileType': 'anndata.zarr', 'url': 'http://localhost:8000/A/0/anndata.zarr',
                                    'options': {
                                        'obsEmbedding': [{'path': 'obsm/X_umap', 'embeddingType': 'UMAP', 'dims': [0, 1]}],
                                        'obsSets': [{'path': 'obs/CellType', 'name': 'Cell Type'}],
                                        'obsLabels': [{'path': 'obs/CellLabel', 'obsLabelsType': 'Cell Label'}]
                                    }})

    def test_anndata_with_base_dir(self):
        adata_path = 'test.h5ad.zarr'
        w = AnnDataWrapper(adata_path, obs_set_paths=['obs/CellType'], obs_set_names=['Cell Type'], obs_embedding_paths=[
                           'obsm/X_umap'], obs_embedding_names=['UMAP'])
        w.base_dir = data_path
        w.local_dir_uid = 'anndata.zarr'

        file_def_creator = w.make_file_def_creator('A', 0)
        file_def = file_def_creator('http://localhost:8000')
        self.assertEqual(file_def, {'fileType': 'anndata.zarr', 'url': 'http://localhost:8000/test.h5ad.zarr',
                                    'options': {
                                        'obsEmbedding': [{'path': 'obsm/X_umap', 'dims': [0, 1], 'embeddingType': 'UMAP'}],
                                        'obsSets': [{'name': 'Cell Type', 'path': 'obs/CellType'}]
                                    }})

    def test_anndata_with_base_dir_no_names(self):
        adata_path = 'test.h5ad.zarr'
        w = AnnDataWrapper(adata_path, obs_set_paths=['obs/CellType'], obs_embedding_paths=[
                           'obsm/X_umap'])
        w.base_dir = data_path
        w.local_dir_uid = 'anndata.zarr'

        file_def_creator = w.make_file_def_creator('A', 0)
        file_def = file_def_creator('http://localhost:8000')
        self.assertEqual(file_def, {'fileType': 'anndata.zarr', 'url': 'http://localhost:8000/test.h5ad.zarr',
                                    'options': {
                                        'obsEmbedding': [{'path': 'obsm/X_umap', 'dims': [0, 1], 'embeddingType': 'X_umap'}],
                                        'obsSets': [{'name': 'CellType', 'path': 'obs/CellType'}]
                                    }})

    def test_anndata_with_h5ad_and_ref_json(self):
        adata_path = data_path / 'test.h5ad'
        ref_json_path = data_path / 'test.h5ad.ref.json'
        w = AnnDataWrapper(adata_path, ref_path=ref_json_path,
                           obs_set_paths=['obs/CellType'], obs_set_names=['Cell Type'],
                           obs_labels_names=['Cell Label'], obs_labels_paths=['obs/CellLabel'],
                           obs_embedding_paths=['obsm/X_umap'], obs_embedding_names=['UMAP'])
        w.local_file_uid = 'anndata.h5ad'
        w.local_ref_uid = 'anndata.reference.json'

        file_def_creator = w.make_file_def_creator('A', 0)
        file_def = file_def_creator('http://localhost:8000')
        self.assertEqual(file_def, {'fileType': 'anndata.h5ad', 'url': 'http://localhost:8000/A/0/anndata.h5ad',
                                    'options': {
                                        'refSpecUrl': 'http://localhost:8000/A/0/anndata.reference.json',
                                        'obsEmbedding': [{'path': 'obsm/X_umap', 'embeddingType': 'UMAP', 'dims': [0, 1]}],
                                        'obsSets': [{'path': 'obs/CellType', 'name': 'Cell Type'}],
                                        'obsLabels': [{'path': 'obs/CellLabel', 'obsLabelsType': 'Cell Label'}]
                                    }})

    def test_anndata_with_h5ad_and_ref_json_with_base_dir(self):
        adata_path = 'test.h5ad'
        ref_json_path = 'test.h5ad.ref.json'
        w = AnnDataWrapper(adata_path, ref_path=ref_json_path,
                           obs_set_paths=['obs/CellType'], obs_set_names=['Cell Type'],
                           obs_labels_names=['Cell Label'], obs_labels_paths=['obs/CellLabel'],
                           obs_embedding_paths=['obsm/X_umap'], obs_embedding_names=['UMAP'])
        w.base_dir = data_path
        w.local_file_uid = 'anndata.h5ad'
        w.local_ref_uid = 'anndata.reference.json'

        file_def_creator = w.make_file_def_creator('A', 0)
        file_def = file_def_creator('http://localhost:8000')
        self.assertEqual(file_def, {'fileType': 'anndata.h5ad', 'url': 'http://localhost:8000/test.h5ad',
                                    'options': {
                                        'refSpecUrl': 'http://localhost:8000/test.h5ad.ref.json',
                                        'obsEmbedding': [{'path': 'obsm/X_umap', 'embeddingType': 'UMAP', 'dims': [0, 1]}],
                                        'obsSets': [{'path': 'obs/CellType', 'name': 'Cell Type'}],
                                        'obsLabels': [{'path': 'obs/CellLabel', 'obsLabelsType': 'Cell Label'}]
                                    }})

    def test_csv(self):
        w = CsvWrapper(
            csv_path=data_path / 'test.umap.csv',
            data_type="obsEmbedding",
            options={
                "obsIndex": "index",
                "obsEmbedding": ["UMAP_1", "UMAP_2"]
            },
            coordination_values={
                "embeddingType": "UMAP"
            }
        )
        w.local_csv_uid = 'test_uid'

        file_def_creator = w.make_csv_file_def_creator(
            "A",
            "0"
        )
        file_def = file_def_creator('http://localhost:8000')
        self.assertEqual(file_def, {
            'fileType': 'obsEmbedding.csv',
            'url': 'http://localhost:8000/A/0/test_uid',
            'options': {
                "obsIndex": "index",
                "obsEmbedding": ["UMAP_1", "UMAP_2"]
            },
            'coordinationValues': {
                "embeddingType": "UMAP"
            }
        })

    def test_csv_with_base_dir(self):
        w = CsvWrapper(
            csv_path='test.umap.csv',
            data_type="obsEmbedding",
            options={
                "obsIndex": "index",
                "obsEmbedding": ["UMAP_1", "UMAP_2"]
            },
            coordination_values={
                "embeddingType": "UMAP"
            }
        )
        w.base_dir = data_path
        w.local_csv_uid = 'test_uid'

        file_def_creator = w.make_csv_file_def_creator(
            "A",
            "0"
        )
        file_def = file_def_creator('http://localhost:8000')
        self.assertEqual(file_def, {
            'fileType': 'obsEmbedding.csv',
            'url': 'http://localhost:8000/test.umap.csv',
            'options': {
                "obsIndex": "index",
                "obsEmbedding": ["UMAP_1", "UMAP_2"]
            },
            'coordinationValues': {
                "embeddingType": "UMAP"
            }
        })

    def test_json(self):
        w = JsonWrapper(
            json_path=data_path / 'test.segmentations.json',
            data_type="obsSegmentations",
            coordination_values={
                "obsType": "nucleus"
            }
        )
        w.local_json_uid = 'test_uid'

        file_def_creator = w.make_json_file_def_creator(
            "A",
            "0"
        )
        file_def = file_def_creator('http://localhost:8000')
        self.assertEqual(file_def, {
            'fileType': 'obsSegmentations.json',
            'url': 'http://localhost:8000/A/0/test_uid',
            'coordinationValues': {
                "obsType": "nucleus"
            }
        })

    def test_json_with_base_dir(self):
        w = JsonWrapper(
            json_path='test.segmentations.json',
            data_type="obsSegmentations",
            coordination_values={
                "obsType": "nucleus"
            }
        )
        w.base_dir = data_path
        w.local_json_uid = 'test_uid'

        file_def_creator = w.make_json_file_def_creator(
            "A",
            "0"
        )
        file_def = file_def_creator('http://localhost:8000')
        self.assertEqual(file_def, {
            'fileType': 'obsSegmentations.json',
            'url': 'http://localhost:8000/test.segmentations.json',
            'coordinationValues': {
                "obsType": "nucleus"
            }
        })

    def test_multivec_zarr(self):
        zarr_filepath = data_path / 'test_out.snap.multivec.zarr'

        w = MultivecZarrWrapper(zarr_path=zarr_filepath)
        w.local_dir_uid = 'test_uid'

        file_def_creator = w.make_genomic_profiles_file_def_creator(
            "A",
            "0"
        )
        file_def = file_def_creator('http://localhost:8000')
        self.assertEqual(file_def, {
            'fileType': 'genomic-profiles.zarr',
            'url': 'http://localhost:8000/A/0/test_uid',
        })

    def test_multivec_zarr_with_base_dir(self):
        zarr_filepath = 'test_out.snap.multivec.zarr'

        w = MultivecZarrWrapper(zarr_path=zarr_filepath)
        w.base_dir = data_path
        w.local_dir_uid = 'test_uid'

        file_def_creator = w.make_genomic_profiles_file_def_creator(
            "A",
            "0"
        )
        file_def = file_def_creator('http://localhost:8000')
        self.assertEqual(file_def, {
            'fileType': 'genomic-profiles.zarr',
            'url': 'http://localhost:8000/test_out.snap.multivec.zarr',
        })

    def test_spatial_data_with_base_dir(self):

        spatial_data_path = 'test.spatialdata.zarr'
        w = SpatialDataWrapper(
            sdata_path=spatial_data_path,
            image_path="images/picture",
            obs_set_paths=['obs/CellType'],
            obs_set_names=['Cell Type'],
            obs_embedding_paths=['obsm/X_umap'],
            obs_embedding_names=['UMAP']
        )
        w.base_dir = data_path
        w.local_dir_uid = 'spatialdata.zarr'

        file_def_creator = w.make_file_def_creator('A', 0)
        file_def = file_def_creator('http://localhost:8000')
        self.assertEqual(file_def, {
            'fileType': 'spatialdata.zarr',
            'url': 'http://localhost:8000/test.spatialdata.zarr',
            'options': {
                'obsSets': {
                    'obsSets': [{'name': 'Cell Type', 'path': 'obs/CellType'}],
                    'tablePath': 'tables/table'
                },
                'image': {'path': 'images/picture'}
            }})

    def test_spatial_data_with_base_dir_2(self):
        spatial_data_path = 'test.spatialdata.zarr'
        w = SpatialDataWrapper(
            sdata_path=spatial_data_path,
            image_path='images/CytAssist_FFPE_Human_Breast_Cancer_full_image',
            coordinate_system='aligned',
            region='CytAssist_FFPE_Human_Breast_Cancer',
            obs_feature_matrix_path='tables/table/X',
            obs_spots_path='shapes/CytAssist_FFPE_Human_Breast_Cancer',
            table_path='tables/table',
            coordination_values={
                "obsType": "spot"
            }
        )
        w.base_dir = data_path
        w.local_dir_uid = 'spatialdata.zarr'

        file_def_creator = w.make_file_def_creator('A', 0)
        file_def = file_def_creator('http://localhost:8000')
        self.assertDictEqual(file_def, {
            'fileType': 'spatialdata.zarr',
            'url': 'http://localhost:8000/test.spatialdata.zarr',
            'options': {
                'image': {
                    'path': 'images/CytAssist_FFPE_Human_Breast_Cancer_full_image',
                    'coordinateSystem': 'aligned',
                },
                'obsFeatureMatrix': {
                    'path': 'tables/table/X',
                    'region': 'CytAssist_FFPE_Human_Breast_Cancer'
                },
                'obsSpots': {
                    'path': 'shapes/CytAssist_FFPE_Human_Breast_Cancer',
                    'tablePath': 'tables/table',
                    'region': 'CytAssist_FFPE_Human_Breast_Cancer',
                    'coordinateSystem': 'aligned',
                }
            },
            'coordinationValues': {
                "obsType": "spot"
            }
        })



import inspect


def make_repr(init_locals, class_def=None):
    '''
    >>> from .wrappers import MultiImageWrapper
    >>> orig = MultiImageWrapper('IMAGE_WRAPPERS', foo='bar')
    >>> orig_repr = repr(orig)
    >>> print(orig_repr)
    MultiImageWrapper(image_wrappers='IMAGE_WRAPPERS', foo='bar')
    >>> evalled = eval(orig_repr)
    >>> assert orig_repr == repr(evalled)
    '''
    # Get the class definition from locals.
    clazz = None
    if '__class__' in init_locals:
        # Requires superclass to be initialized.
        clazz = init_locals.pop('__class__')
    elif 'self' in init_locals and hasattr(init_locals['self'], '__class__'):
        clazz = init_locals["self"].__class__  # pragma: no cover
    elif class_def is not None:
        clazz = class_def
    else:
        raise ValueError("make_repr could not locate the class definition")  # pragma: no cover

    # Remove self from locals.
    if 'self' in init_locals:
        del init_locals['self']

    # Get the class name.
    class_name = clazz.__name__

    # Remove redundant constructor parameters (when the value equals the default value).
    for k, v in inspect.signature(clazz).parameters.items():
        if k not in init_locals:
            continue

        try:
            is_default = init_locals[k] == v.default
        except NotImplementedError:
            # Equality comparison may not be implemented for the value object.
            is_default = False

        try:
            if is_default:
                del init_locals[k]
        except ValueError:
            # TODO: Is this also expected?
            # ValueError('The truth value of a DataFrame is ambiguous. Use a.empty, a.bool(), a.item(), a.any() or a.all().')
            pass

    # Convert the kwargs dict to named args.
    if 'kwargs' in init_locals:
        kwargs = init_locals.pop('kwargs')
    else:
        kwargs = {}

    args = {
        **init_locals,
        **kwargs
    }
    params = ', '.join([f'{k}={repr(v)}' for k, v in args.items()])
    return f'{class_name}({params})'


def make_params_repr(args):
    '''
    >>> print(make_params_repr({ "uid": 1, "name": "My Dataset"}))
    uid=1, name='My Dataset'
    '''
    params = ', '.join([f'{k}={repr(v)}' for k, v in args.items()])
    return params



import numpy as np
from anndata import AnnData
import scipy.cluster
from scipy.sparse import issparse

VAR_CHUNK_SIZE = 10


def generate_h5ad_ref_spec(h5_url, omit_url=True):
    from kerchunk.hdf import SingleHdf5ToZarr
    h5chunks = SingleHdf5ToZarr(h5_url, inline_threshold=300)
    h5dict = h5chunks.translate()
    if omit_url:
        for key, val in h5dict['refs'].items():
            if isinstance(val, list):
                h5dict['refs'][key] = [None, *val[1:]]
    return h5dict


def cast_arr(arr):
    """
    Try to cast an array to a dtype that takes up less space.

    :param arr: The array to cast.
    :type arr: np.array

    :returns: The new array.
    :rtype: np.array
    """
    orig_sum = np.sum(arr)
    orig_max = np.max(arr)
    orig_min = np.min(arr)

    # Try casting float to int for better downstream compression.
    if arr.dtype.kind == 'f':
        cast_arr = arr.astype(f'<i{arr.dtype.itemsize}')
        cast_sum = np.sum(cast_arr)
        if np.abs(orig_sum - cast_sum) < 1e-2:
            arr = cast_arr

    # Try casting signed int to unsigned int.
    if arr.dtype.kind == 'i':
        # If signed int dtype but no values are less than zero,
        # convert to unsigned int dtype.
        if arr.min() >= 0:
            arr = arr.astype(f'<u{arr.dtype.itemsize}')

    # Try casting to a smaller itemsize.
    if arr.dtype.kind == 'u' or arr.dtype.kind == 'i' or arr.dtype.kind == 'f':
        next_itemsizes = [4] if arr.dtype.kind == 'f' else [4, 2, 1]
        for next_itemsize in next_itemsizes:
            if arr.dtype.itemsize > next_itemsize:
                next_dtype = np.dtype(f'<{arr.dtype.kind}{next_itemsize}')
                next_dtype_info = np.iinfo(next_dtype) if arr.dtype.kind == 'u' or arr.dtype.kind == 'i' else np.finfo(next_dtype)
                if next_dtype_info.min <= orig_min and next_dtype_info.max >= orig_max:
                    arr = arr.astype(next_dtype)
                elif arr.dtype.itemsize == 8 and (arr.dtype.kind == 'u' or arr.dtype.kind == 'i'):
                    print(f"WARNING: Not casting array with dtype {arr.dtype.name}, but Zarr.js suggests avoiding int64 and uint64")

    # Check for float16 usage.
    if arr.dtype.kind == 'f' and arr.dtype.itemsize == 2:
        # Zarr.js does not have a Float16Array type
        arr = arr.astype('<f4')

    return arr


def optimize_arr(arr):
    """
    Try to cast an array to a dtype that takes up less space, and convert to dense.

    :param arr: The array to cast and convert.
    :type arr: np.array

    :returns: The new array.
    :rtype: np.array
    """
    arr = to_dense(cast_arr(to_memory(arr)))
    if isinstance(arr, np.matrix):
        arr = np.array(arr)
    return arr


def optimize_adata(adata, obs_cols=None, obsm_keys=None, var_cols=None, varm_keys=None, layer_keys=None, remove_X=False, optimize_X=False, to_dense_X=False, to_sparse_X=False):
    """
    Given an AnnData object, optimize for usage with Vitessce and return a new object.

    :param adata: The AnnData object to optimize.
    :type adata: anndata.AnnData
    :param obs_cols: Columns of adata.obs to optimize. Columns not specified will not be included in the returned object.
    :type obs_cols: list[str] or None
    :param var_cols: Columns of adata.var to optimize. Columns not specified will not be included in the returned object.
    :type var_cols: list[str] or None
    :param obsm_keys: Arrays within adata.obsm to optimize. Keys not specified will not be included in the returned object.
    :type obsm_keys: list[str] or None
    :param varm_keys: Arrays within adata.varm to optimize. Keys not specified will not be included in the returned object.
    :type varm_keys: list[str] or None
    :param layer_keys: Arrays within adata.layers to optimize. Keys not specified will not be included in the returned object.
    :type layer_keys: list[str] or None
    :param bool remove_X: Should the returned object have its X matrix set to None? By default, False.
    :param bool optimize_X: Should the returned object run optimize_arr on adata.X? By default, False.
    :param bool to_dense_X: Should adata.X be cast to a dense array in the returned object? By default, False.
    :param bool to_sparse_X: Should adata.X be cast to a sparse array in the returned object? By default, False.

    :returns: The new AnnData object.
    :rtype: anndata.AnnData
    """
    if not remove_X:
        if adata.X is not None:
            if optimize_X:
                new_X = optimize_arr(adata.X)
            else:
                new_X = adata.X
            # Convert to dense or sparse if asked
            if to_dense_X:
                new_X = to_dense(new_X)
            if to_sparse_X:
                new_X = new_X.tocsc()
            if to_dense_X and to_sparse_X:
                raise ValueError(
                    "Did not expect both to_dense_X and to_sparse_X to be True")
            # If the user has not asked for their matrix to be converted to dense or sparse,
            # we still want to ensure that, if the original matrix was sparse, it
            # uses the CSC sparse format.
            if not to_dense_X and not to_sparse_X:
                # In the future, we can use sparse matrices with equal performance:
                # https://github.com/theislab/anndata/issues/524
                # Vitessce can use csc matrices somewhat efficiently.
                if issparse(new_X):
                    new_X = new_X.tocsc()
        else:
            new_X = None
    else:
        new_X = None
    # Only keep the subset of columns required.
    if obs_cols is not None:
        new_obs = adata.obs[obs_cols]
    else:
        new_obs = adata.obs
    if var_cols is not None:
        new_var = adata.var[var_cols]
    else:
        new_var = adata.var
    # Only keep the subset of obsm and varm items required.
    if obsm_keys is None:
        obsm_keys = adata.obsm.keys()
    new_obsm = {
        obsm_key: optimize_arr(adata.obsm[obsm_key])
        for obsm_key
        in obsm_keys
    }
    if varm_keys is None:
        varm_keys = adata.varm.keys()
    new_varm = {
        varm_key: optimize_arr(adata.varm[varm_key])
        for varm_key
        in varm_keys
    }
    if layer_keys is None:
        layer_keys = adata.layers.keys()
    new_layers = {
        layer_key: optimize_arr(adata.layers[layer_key])
        for layer_key
        in layer_keys
    }
    adata = AnnData(X=new_X, obs=new_obs, var=new_var, obsm=new_obsm, varm=new_varm, layers=new_layers)
    return adata


def to_memory(arr):
    """
    Try to load a backed AnnData array into memory.

    :param arr: The array to load.
    :type arr: np.array

    :returns: The loaded array.
    :rtype: np.array
    """
    if type(arr).__name__ == "SparseDataset":
        # This comes from using a backed AnnData object (e.g. read_h5ad(path, backed="r+")).
        # Reference: https://github.com/scverse/anndata/blob/286bc7f207863964cb861f17c96ab24fe0cf72ac/anndata/_core/sparse_dataset.py#L230
        return arr.to_memory()
    return arr


def to_dense(arr):
    """
    Convert a sparse array to dense.

    :param arr: The array to convert.
    :type arr: np.array

    :returns: The converted array (or the original array if it was already dense).
    :rtype: np.array
    """
    if issparse(arr):
        return arr.todense()
    return arr


def to_uint8(arr, norm_along=None):
    """
    Convert an array to uint8 dtype.

    :param arr: The array to convert.
    :type arr: np.array
    :param norm_along: How to normalize the array values. By default, None. Valid values are "global", "var", "obs".
    :type norm_along: str or None

    :returns: The converted array.
    :rtype: np.array
    """
    # Re-scale the gene expression values between 0 and 255 (one byte ints).
    if norm_along is None:
        norm_arr = arr
    elif norm_along == "global":
        arr *= 255.0 / arr.max()
        norm_arr = arr
    elif norm_along == "var":
        # Normalize along gene axis
        arr = to_dense(arr)
        num_cells = arr.shape[0]
        min_along_genes = arr.min(axis=0)
        max_along_genes = arr.max(axis=0)
        range_per_gene = max_along_genes - min_along_genes
        ratio_per_gene = 255.0 / range_per_gene

        norm_arr = np.multiply(
            (arr - np.tile(min_along_genes, (num_cells, 1))),
            np.tile(ratio_per_gene, (num_cells, 1))
        )
    elif norm_along == "obs":
        # Normalize along cell axis
        arr = to_dense(arr)
        num_genes = arr.shape[1]
        min_along_cells = arr.min(axis=1)
        max_along_cells = arr.max(axis=1)
        range_per_cell = max_along_cells - min_along_cells
        ratio_per_cell = 255.0 / range_per_cell

        norm_arr = np.multiply(
            (arr.T - np.tile(min_along_cells, (num_genes, 1))),
            np.tile(ratio_per_cell, (num_genes, 1))
        ).T
    else:
        raise ValueError("to_uint8 received unknown norm_along value")
    return norm_arr.astype('u1')


def sort_var_axis(adata_X, orig_var_index, full_var_index=None):
    """
    Sort the var index by performing hierarchical clustering.

    :param adata_X: The matrix to use for clustering. For example, adata.X
    :type adata_X: np.array
    :param orig_var_index: The original var index. For example, adata.var.index
    :type orig_var_index: pandas.Index
    :param full_var_index: Pass the full adata.var.index to append the var values excluded from sorting, if adata_X and orig_var_index are a subset of the full adata.X matrix. By default, None.
    :type full_var_index: pandas.Index or None

    :returns: The sorted elements of the var index.
    :rtype: list[str]
    """
    gexp_arr = to_dense(adata_X)

    # Perform hierarchical clustering along the genes axis.
    Z = scipy.cluster.hierarchy.linkage(gexp_arr.T, method="ward")

    # Get the hierarchy-based ordering of genes.
    leaf_index_list = scipy.cluster.hierarchy.leaves_list(Z)
    leaf_list = orig_var_index[leaf_index_list].tolist()

    if full_var_index:
        leaf_list = leaf_list + list(set(full_var_index).difference(set(leaf_list)))

    return leaf_list


def to_diamond(x, y, r):
    """
    Convert an (x, y) coordinate to a polygon (diamond) with a given radius.

    :param x: The x coordinate.
    :type x: int or float
    :param y: The y coordinate.
    :type x: int or float
    :param r: The radius.
    :type r: int or float

    :returns: The polygon vertices as an array of coordinate pairs, like [[x1, y1], [x2, y2], ...]
    :rtype: np.array
    """
    return np.array([[x, y + r], [x + r, y], [x, y - r], [x - r, y]])



from .anndata import (
    optimize_adata,
    optimize_arr,
    to_dense,
    to_uint8,
    sort_var_axis,
    to_diamond,
    VAR_CHUNK_SIZE,
    generate_h5ad_ref_spec,
)
from .ome import (
    rgb_img_to_ome_zarr,
    multiplex_img_to_ome_zarr,
    rgb_img_to_ome_tiff,
    multiplex_img_to_ome_tiff,
)
from .multivec import (
    adata_to_multivec_zarr,
)
